@@ -78,6 +78,10 @@
  * Use [method@Dialog.force_close] to close the dialog even when `can-close` is set to
  * `FALSE`.
  *
+ * `AdwDialog` is transient and doesn't integrate with the window below it, for
+ * example it's not possible to collapse it into a bottom bar. See
+ * [class@BottomSheet] for persistent and more tightly integrated bottom sheets.
+ *
  * ## Header Bar Integration
  *
  * When placed inside an `AdwDialog`, [class@HeaderBar] will display the dialog
@@ -128,6 +132,7 @@ typedef struct
   AdwDialogPresentationMode presentation_mode;
   AdwBreakpoint *portrait_breakpoint;
   AdwBreakpoint *landscape_breakpoint;
+  AdwBreakpoint *both_breakpoint;
 
   GtkWidget *focus_widget;
   GtkWidget *default_widget;
@@ -202,7 +207,8 @@ map_tick_cb (AdwDialog *self)
 }
 
 static void
-sheet_closing_cb (AdwDialog *self)
+sheet_closing_cb (GtkWidget *sheet,
+                  AdwDialog *self)
 {
   AdwDialogPrivate *priv = adw_dialog_get_instance_private (self);
 
@@ -213,7 +219,8 @@ sheet_closing_cb (AdwDialog *self)
 }
 
 static void
-sheet_closed_cb (AdwDialog *self)
+sheet_closed_cb (GtkWidget *sheet,
+                 AdwDialog *self)
 {
   AdwDialogPrivate *priv = adw_dialog_get_instance_private (self);
 
@@ -224,7 +231,6 @@ sheet_closed_cb (AdwDialog *self)
 static void
 sheet_close_attempt_cb (AdwDialog *self)
 {
-  g_print ("a\n");
   g_signal_emit (self, signals[SIGNAL_CLOSE_ATTEMPT], 0);
 }
 
@@ -436,7 +442,27 @@ update_presentation (AdwDialog *self)
   breakpoint =
     adw_breakpoint_bin_get_current_breakpoint (ADW_BREAKPOINT_BIN (priv->bin));
   use_bottom_sheet = priv->presentation_mode == ADW_DIALOG_BOTTOM_SHEET ||
-                     breakpoint != NULL;
+                     (priv->presentation_mode == ADW_DIALOG_AUTO && breakpoint != NULL);
+
+  if (breakpoint == priv->both_breakpoint) {
+    gtk_widget_add_css_class (GTK_WIDGET (self), "narrow");
+    gtk_widget_add_css_class (GTK_WIDGET (self), "short");
+  } else if (breakpoint == priv->portrait_breakpoint) {
+    gtk_widget_add_css_class (GTK_WIDGET (self), "narrow");
+    gtk_widget_remove_css_class (GTK_WIDGET (self), "short");
+  } else if (breakpoint == priv->landscape_breakpoint) {
+    gtk_widget_remove_css_class (GTK_WIDGET (self), "narrow");
+    gtk_widget_add_css_class (GTK_WIDGET (self), "short");
+  } else {
+    gtk_widget_remove_css_class (GTK_WIDGET (self), "narrow");
+    gtk_widget_remove_css_class (GTK_WIDGET (self), "short");
+  }
+
+  if (use_bottom_sheet && priv->bottom_sheet)
+    return;
+
+  if (!use_bottom_sheet && priv->floating_sheet)
+    return;
 
   g_object_ref (priv->child_breakpoint_bin);
 
@@ -450,11 +476,6 @@ update_presentation (AdwDialog *self)
 
     if (focus)
       g_object_add_weak_pointer (G_OBJECT (focus), (gpointer *) &focus);
-
-    /* FIXME: the focused widget has a strong ref and it never gets cleared. The
-     * real source of this leak is likely somewhere in GTK, but unsetting the
-     * focus "fixes" it */
-    gtk_root_set_focus (root, NULL);
   }
 
   if (priv->bottom_sheet) {
@@ -475,29 +496,23 @@ update_presentation (AdwDialog *self)
     if (!priv->first_map)
       adw_bottom_sheet_set_open (priv->bottom_sheet, TRUE);
 
+    adw_bottom_sheet_set_full_width (priv->bottom_sheet, FALSE);
     adw_bottom_sheet_set_show_drag_handle (priv->bottom_sheet, FALSE);
     adw_bottom_sheet_set_sheet (priv->bottom_sheet, priv->child_breakpoint_bin);
     adw_bottom_sheet_set_can_close (priv->bottom_sheet, priv->can_close);
     adw_breakpoint_bin_set_child (ADW_BREAKPOINT_BIN (priv->bin),
                                   GTK_WIDGET (priv->bottom_sheet));
 
-    g_signal_connect_swapped (priv->bottom_sheet, "closing",
-                              G_CALLBACK (sheet_closing_cb), self);
-    g_signal_connect_swapped (priv->bottom_sheet, "closed",
-                              G_CALLBACK (sheet_closed_cb), self);
+    adw_bottom_sheet_set_callbacks (priv->bottom_sheet,
+                                    (GFunc) sheet_closing_cb,
+                                    (GFunc) sheet_closed_cb,
+                                    self);
+
     g_signal_connect_swapped (priv->bottom_sheet, "close-attempt",
                               G_CALLBACK (sheet_close_attempt_cb), self);
 
     gtk_widget_add_css_class (GTK_WIDGET (self), "bottom-sheet");
     gtk_widget_remove_css_class (GTK_WIDGET (self), "floating");
-
-    if (breakpoint == priv->portrait_breakpoint) {
-      gtk_widget_add_css_class (GTK_WIDGET (self), "portrait");
-      gtk_widget_remove_css_class (GTK_WIDGET (self), "landscape");
-    } else {
-      gtk_widget_add_css_class (GTK_WIDGET (self), "landscape");
-      gtk_widget_remove_css_class (GTK_WIDGET (self), "portrait");
-    }
   } else {
     priv->floating_sheet = ADW_FLOATING_SHEET (adw_floating_sheet_new ());
 
@@ -509,10 +524,11 @@ update_presentation (AdwDialog *self)
     adw_breakpoint_bin_set_child (ADW_BREAKPOINT_BIN (priv->bin),
                                   GTK_WIDGET (priv->floating_sheet));
 
-    g_signal_connect_swapped (priv->floating_sheet, "closing",
-                              G_CALLBACK (sheet_closing_cb), self);
-    g_signal_connect_swapped (priv->floating_sheet, "closed",
-                              G_CALLBACK (sheet_closed_cb), self);
+    adw_floating_sheet_set_callbacks (priv->floating_sheet,
+                                      (GFunc) sheet_closing_cb,
+                                      (GFunc) sheet_closed_cb,
+                                      self);
+
     g_signal_connect_swapped (priv->floating_sheet, "close-attempt",
                               G_CALLBACK (sheet_close_attempt_cb), self);
 
@@ -524,7 +540,7 @@ update_presentation (AdwDialog *self)
 
   if (focus) {
     gtk_widget_grab_focus (focus);
-    g_object_remove_weak_pointer (G_OBJECT (focus), (gpointer *) &focus);
+    g_clear_weak_pointer (&focus);
   }
 
   g_object_unref (priv->child_breakpoint_bin);
@@ -535,40 +551,10 @@ update_presentation_mode (AdwDialog *self)
 {
   AdwDialogPrivate *priv = adw_dialog_get_instance_private (self);
 
-  if (priv->window)
+  if (priv->window || !priv->bin)
     return;
 
-  switch (priv->presentation_mode) {
-  case ADW_DIALOG_AUTO:
-    g_assert (!priv->portrait_breakpoint);
-    g_assert (!priv->landscape_breakpoint);
-
-    priv->landscape_breakpoint = adw_breakpoint_new (adw_breakpoint_condition_parse (LANDSCAPE_CONDITION));
-    adw_breakpoint_bin_add_breakpoint (ADW_BREAKPOINT_BIN (priv->bin),
-                                       priv->landscape_breakpoint);
-
-    priv->portrait_breakpoint = adw_breakpoint_new (adw_breakpoint_condition_parse (PORTRAIT_CONDITION));
-    adw_breakpoint_bin_add_breakpoint (ADW_BREAKPOINT_BIN (priv->bin),
-                                       priv->portrait_breakpoint);
-    break;
-  case ADW_DIALOG_FLOATING:
-  case ADW_DIALOG_BOTTOM_SHEET:
-    if (priv->portrait_breakpoint) {
-      adw_breakpoint_bin_remove_breakpoint (ADW_BREAKPOINT_BIN (priv->bin),
-                                            priv->portrait_breakpoint);
-      priv->portrait_breakpoint = NULL;
-
-      adw_breakpoint_bin_remove_breakpoint (ADW_BREAKPOINT_BIN (priv->bin),
-                                            priv->landscape_breakpoint);
-      priv->landscape_breakpoint = NULL;
-    }
-    break;
-  default:
-    g_assert_not_reached ();
-  }
-
-  if (priv->bin)
-    update_presentation (self);
+  update_presentation (self);
 }
 
 static void
@@ -1045,7 +1031,7 @@ adw_dialog_class_init (AdwDialogClass *klass)
   widget_class->compute_expand = adw_widget_compute_expand;
 
   /**
-   * AdwDialog:child: (attributes org.gtk.Property.get=adw_dialog_get_child org.gtk.Property.set=adw_dialog_set_child)
+   * AdwDialog:child:
    *
    * The child widget of the `AdwDialog`.
    *
@@ -1057,7 +1043,7 @@ adw_dialog_class_init (AdwDialogClass *klass)
                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwDialog:title: (attributes org.gtk.Property.get=adw_dialog_get_title org.gtk.Property.set=adw_dialog_set_title)
+   * AdwDialog:title:
    *
    * The title of the dialog.
    *
@@ -1069,7 +1055,7 @@ adw_dialog_class_init (AdwDialogClass *klass)
                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwDialog:can-close: (attributes org.gtk.Property.get=adw_dialog_get_can_close org.gtk.Property.set=adw_dialog_set_can_close)
+   * AdwDialog:can-close:
    *
    * Whether the dialog can be closed.
    *
@@ -1086,7 +1072,7 @@ adw_dialog_class_init (AdwDialogClass *klass)
                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwDialog:content-width: (attributes org.gtk.Property.get=adw_dialog_get_content_width org.gtk.Property.set=adw_dialog_set_content_width)
+   * AdwDialog:content-width:
    *
    * The width of the dialog's contents.
    *
@@ -1102,7 +1088,7 @@ adw_dialog_class_init (AdwDialogClass *klass)
                       G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwDialog:content-height: (attributes org.gtk.Property.get=adw_dialog_get_content_height org.gtk.Property.set=adw_dialog_set_content_height)
+   * AdwDialog:content-height:
    *
    * The height of the dialog's contents.
    *
@@ -1118,7 +1104,7 @@ adw_dialog_class_init (AdwDialogClass *klass)
                       G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwDialog:follows-content-size: (attributes org.gtk.Property.get=adw_dialog_get_follows_content_size org.gtk.Property.set=adw_dialog_set_follows_content_size)
+   * AdwDialog:follows-content-size:
    *
    * Whether to size content automatically.
    *
@@ -1136,7 +1122,7 @@ adw_dialog_class_init (AdwDialogClass *klass)
                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwDialog:presentation-mode: (attributes org.gtk.Property.get=adw_dialog_get_presentation_mode org.gtk.Property.set=adw_dialog_set_presentation_mode)
+   * AdwDialog:presentation-mode:
    *
    * The dialog's presentation mode.
    *
@@ -1159,7 +1145,7 @@ adw_dialog_class_init (AdwDialogClass *klass)
                        G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwDialog:focus-widget: (attributes org.gtk.Property.get=adw_dialog_get_focus org.gtk.Property.set=adw_dialog_set_focus)
+   * AdwDialog:focus-widget: (getter get_focus) (setter set_focus):
    *
    * The focus widget.
    *
@@ -1171,7 +1157,7 @@ adw_dialog_class_init (AdwDialogClass *klass)
                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwDialog:default-widget: (attributes org.gtk.Property.get=adw_dialog_get_default_widget org.gtk.Property.set=adw_dialog_set_default_widget)
+   * AdwDialog:default-widget:
    *
    * The default widget.
    *
@@ -1185,7 +1171,7 @@ adw_dialog_class_init (AdwDialogClass *klass)
                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwDialog:current-breakpoint: (attributes org.gtk.Property.get=adw_dialog_get_current_breakpoint)
+   * AdwDialog:current-breakpoint:
    *
    * The current breakpoint.
    *
@@ -1272,9 +1258,6 @@ adw_dialog_class_init (AdwDialogClass *klass)
   gtk_widget_class_add_binding (widget_class, GDK_KEY_D, GDK_CONTROL_MASK | GDK_SHIFT_MASK,
                                 (GtkShortcutFunc) open_inspector_cb, "b", TRUE);
 
-  gtk_widget_class_add_binding (widget_class, GDK_KEY_Escape, 0,
-                                (GtkShortcutFunc) adw_dialog_close, NULL);
-
   gtk_widget_class_set_layout_manager_type (widget_class, GTK_TYPE_BIN_LAYOUT);
   gtk_widget_class_set_css_name (widget_class, "dialog");
   gtk_widget_class_set_accessible_role (widget_class, GTK_ACCESSIBLE_ROLE_DIALOG);
@@ -1294,6 +1277,7 @@ adw_dialog_init (AdwDialog *self)
   priv->presentation_mode = ADW_DIALOG_AUTO;
 
   priv->child_breakpoint_bin = adw_breakpoint_bin_new ();
+  gtk_widget_set_overflow (priv->child_breakpoint_bin, GTK_OVERFLOW_VISIBLE);
   adw_breakpoint_bin_set_warning_widget (ADW_BREAKPOINT_BIN (priv->child_breakpoint_bin),
                                          GTK_WIDGET (self));
   g_object_bind_property (self, "width-request",
@@ -1343,7 +1327,7 @@ adw_dialog_new (void)
 }
 
 /**
- * adw_dialog_get_child: (attributes org.gtk.Method.get_property=child)
+ * adw_dialog_get_child:
  * @self: a dialog
  *
  * Gets the child widget of @self.
@@ -1365,7 +1349,7 @@ adw_dialog_get_child (AdwDialog *self)
 }
 
 /**
- * adw_dialog_set_child: (attributes org.gtk.Method.set_property=child)
+ * adw_dialog_set_child:
  * @self: a dialog
  * @child: (nullable): the child widget
  *
@@ -1398,7 +1382,7 @@ adw_dialog_set_child (AdwDialog *self,
 }
 
 /**
- * adw_dialog_get_title: (attributes org.gtk.Method.get_property=title)
+ * adw_dialog_get_title:
  * @self: a dialog
  *
  * Gets the title of @self.
@@ -1420,7 +1404,7 @@ adw_dialog_get_title (AdwDialog *self)
 }
 
 /**
- * adw_dialog_set_title: (attributes org.gtk.Method.set_property=title)
+ * adw_dialog_set_title:
  * @self: a dialog
  * @title: (transfer none): the new title
  *
@@ -1449,7 +1433,7 @@ adw_dialog_set_title (AdwDialog  *self,
 }
 
 /**
- * adw_dialog_get_can_close: (attributes org.gtk.Method.get_property=can-close)
+ * adw_dialog_get_can_close:
  * @self: a dialog
  *
  * Gets whether @self can be closed.
@@ -1471,7 +1455,7 @@ adw_dialog_get_can_close (AdwDialog *self)
 }
 
 /**
- * adw_dialog_set_can_close: (attributes org.gtk.Method.set_property=can-close)
+ * adw_dialog_set_can_close:
  * @self: a dialog
  * @can_close: whether to allow closing
  *
@@ -1511,7 +1495,7 @@ adw_dialog_set_can_close (AdwDialog *self,
 }
 
 /**
- * adw_dialog_get_content_width: (attributes org.gtk.Method.get_property=content-width)
+ * adw_dialog_get_content_width:
  * @self: a dialog
  *
  * Gets the width of the dialog's contents.
@@ -1533,7 +1517,7 @@ adw_dialog_get_content_width (AdwDialog *self)
 }
 
 /**
- * adw_dialog_set_content_width: (attributes org.gtk.Method.set_property=content-width)
+ * adw_dialog_set_content_width:
  * @self: a dialog
  * @content_width: the content width
  *
@@ -1561,7 +1545,7 @@ adw_dialog_set_content_width (AdwDialog *self,
 }
 
 /**
- * adw_dialog_get_content_height: (attributes org.gtk.Method.get_property=content-height)
+ * adw_dialog_get_content_height:
  * @self: a dialog
  *
  * Gets the height of the dialog's contents.
@@ -1583,7 +1567,7 @@ adw_dialog_get_content_height (AdwDialog *self)
 }
 
 /**
- * adw_dialog_set_content_height: (attributes org.gtk.Method.set_property=content-height)
+ * adw_dialog_set_content_height:
  * @self: a dialog
  * @content_height: the content height
  *
@@ -1611,7 +1595,7 @@ adw_dialog_set_content_height (AdwDialog *self,
 }
 
 /**
- * adw_dialog_get_follows_content_size: (attributes org.gtk.Method.get_property=follows-content-size)
+ * adw_dialog_get_follows_content_size:
  * @self: a dialog
  *
  * Gets whether to size content of @self automatically.
@@ -1633,7 +1617,7 @@ adw_dialog_get_follows_content_size (AdwDialog *self)
 }
 
 /**
- * adw_dialog_set_follows_content_size: (attributes org.gtk.Method.set_property=follows-content-size)
+ * adw_dialog_set_follows_content_size:
  * @self: a dialog
  * @follows_content_size: whether to size content automatically
  *
@@ -1670,7 +1654,7 @@ adw_dialog_set_follows_content_size (AdwDialog *self,
 }
 
 /**
- * adw_dialog_get_presentation_mode: (attributes org.gtk.Method.get_property=presentation-mode)
+ * adw_dialog_get_presentation_mode:
  * @self: a dialog
  *
  * Gets presentation mode for @self.
@@ -1692,7 +1676,7 @@ adw_dialog_get_presentation_mode (AdwDialog *self)
 }
 
 /**
- * adw_dialog_set_presentation_mode: (attributes org.gtk.Method.set_property=presentation-mode)
+ * adw_dialog_set_presentation_mode:
  * @self: a dialog
  * @presentation_mode: the new presentation mode
  *
@@ -1733,7 +1717,7 @@ adw_dialog_set_presentation_mode (AdwDialog                 *self,
 }
 
 /**
- * adw_dialog_get_focus: (attributes org.gtk.Method.get_property=focus-widget)
+ * adw_dialog_get_focus: (get-property focus-widget)
  * @self: a dialog
  *
  * Gets the focus widget for @self.
@@ -1755,7 +1739,7 @@ adw_dialog_get_focus (AdwDialog *self)
 }
 
 /**
- * adw_dialog_set_focus: (attributes org.gtk.Method.set_property=focus-widget)
+ * adw_dialog_set_focus: (set-property focus-widget)
  * @self: a dialog
  * @focus: (nullable) (transfer none): the focus widget
  *
@@ -1805,7 +1789,7 @@ adw_dialog_set_focus (AdwDialog *self,
 }
 
 /**
- * adw_dialog_get_default_widget: (attributes org.gtk.Method.get_property=default-widget)
+ * adw_dialog_get_default_widget:
  * @self: a dialog
  *
  * Gets the default widget for @self.
@@ -1827,7 +1811,7 @@ adw_dialog_get_default_widget (AdwDialog *self)
 }
 
 /**
- * adw_dialog_set_default_widget: (attributes org.gtk.Method.set_property=default-widget)
+ * adw_dialog_set_default_widget:
  * @self: a dialog
  * @default_widget: (nullable) (transfer none): the default widget
  *
@@ -1911,7 +1895,6 @@ adw_dialog_close (AdwDialog *self)
   priv = adw_dialog_get_instance_private (self);
 
   if (!priv->can_close) {
-  g_print ("b\n");
     g_signal_emit (self, signals[SIGNAL_CLOSE_ATTEMPT], 0);
     return FALSE;
   }
@@ -1984,7 +1967,7 @@ adw_dialog_add_breakpoint (AdwDialog     *self,
 }
 
 /**
- * adw_dialog_get_current_breakpoint: (attributes org.gtk.Method.get_property=current-breakpoint)
+ * adw_dialog_get_current_breakpoint:
  * @self: a dialog
  *
  * Gets the current breakpoint.
@@ -2088,6 +2071,14 @@ adw_dialog_present (AdwDialog *self,
     adw_breakpoint_bin_set_pass_through (ADW_BREAKPOINT_BIN (priv->bin), TRUE);
     adw_breakpoint_bin_set_warnings (ADW_BREAKPOINT_BIN (priv->bin), FALSE, TRUE);
 
+    priv->landscape_breakpoint = adw_breakpoint_new (adw_breakpoint_condition_parse (LANDSCAPE_CONDITION));
+    priv->portrait_breakpoint = adw_breakpoint_new (adw_breakpoint_condition_parse (PORTRAIT_CONDITION));
+    priv->both_breakpoint = adw_breakpoint_new (adw_breakpoint_condition_parse (PORTRAIT_CONDITION " and " LANDSCAPE_CONDITION));
+
+    adw_breakpoint_bin_add_breakpoint (ADW_BREAKPOINT_BIN (priv->bin), priv->landscape_breakpoint);
+    adw_breakpoint_bin_add_breakpoint (ADW_BREAKPOINT_BIN (priv->bin), priv->portrait_breakpoint);
+    adw_breakpoint_bin_add_breakpoint (ADW_BREAKPOINT_BIN (priv->bin), priv->both_breakpoint);
+
     gtk_widget_set_parent (priv->bin, GTK_WIDGET (self));
 
     g_signal_connect_swapped (priv->bin, "notify::current-breakpoint",
